#include <stdio.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>

struct msh_head
{
	int type;
	char name[16];
	int unk;
	unsigned int faces;
	unsigned int verts;
  int var[18];
};

struct vector3f
{
	float x,y,z; 
};

struct vector2f
{
	float tx,ty; 
};

struct face
{
	unsigned short f0,f1,f2; 
};

struct msh_bbox {
	struct vector3f		min;
	struct vector3f		max;
};

int main(int argc, char* argv[])
{
  msh_head hdr;
  unsigned long i;
	char fileName[256];
  FILE *inputFile;

	printf("Orion MSH parser [201101xx]\n");
	if (argc < 2) 
  {
		printf("usage: parse_msh < srcfile.msh >\n");
		return 1;
	}

	strcpy((char*)fileName,argv[1]);

  inputFile = fopen(fileName,"rb");
	if(!inputFile)
  {
 	printf("ERR: cannot open.\n");
  return 1;
  }

	fseek(inputFile,0,SEEK_SET);

  fread( &hdr, sizeof(msh_head), 1, inputFile );
  printf("Type 0x%04X\nVersion '%s'\n",hdr.type,hdr.name);

  printf("Unk 0x%08X (%d)\n",hdr.unk,hdr.unk);
  printf("Faces 0x%08X (%d)\n",hdr.faces,hdr.faces);
  printf("Verts 0x%08X (%d)\n",hdr.verts,hdr.verts);
  for (i=0;i<18;i++)
  printf("Var 0x%08X (%d)\n",hdr.var[i],hdr.var[i]);

  msh_bbox * bb = new msh_bbox;
	fseek(inputFile,0x50,SEEK_SET);
  fread(bb, sizeof(msh_bbox), 1, inputFile );
  printf("bb_min %f %f %f bb_max %f %f %f \n",
   bb->min.x,bb->min.y,bb->min.z,bb->max.x,bb->max.y,bb->max.z);


  vector3f * v = new vector3f[hdr.verts];
	fseek(inputFile,0x68,SEEK_SET);

  fread(v, sizeof(vector3f)*hdr.verts, 1, inputFile );
  for (i=0;i<hdr.verts;i++)
  printf("v %f %f %f\n",v[i].x,v[i].y,v[i].z);

  fread(v, sizeof(vector3f)*hdr.verts, 1, inputFile );
  for (i=0;i<hdr.verts;i++)
  printf("vn %f %f %f\n",v[i].x,v[i].y,v[i].z);
  delete [] v;

  face * f = new face[hdr.faces];
  fread(f, sizeof(face)*hdr.faces, 1, inputFile );
  for (i=0;i<hdr.faces;i++)
  printf("f %d/%d/%d\n",f[i].f0,f[i].f1,f[i].f2);
  delete [] f;

  vector2f * vt = new vector2f[hdr.verts];
  fread(vt, sizeof(vector2f)*hdr.verts, 1, inputFile );
  for (i=0;i<hdr.verts;i++)
  printf("vt %f %f\n",vt[i].tx,vt[i].ty);
  delete [] vt;

	//printf("Full size is 0x%08X (%u)\n",fullsize,fullsize);

  fclose(inputFile);
	return 0;
	}